"""
{
    "fields":
        [
            {
                "name"
                "type"
            }
        ],
    "steps": {
        "id": <uuid>,
        "operator": "name",
        "columns": [],
        "config": {<variable>},
        "requires": [<uuid>]
    }
}

"""
